"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.error = exports.middleware = exports.resolvers = void 0;
const ono_1 = require("ono");
const _uniq = require("lodash.uniq");
const middlewares = require("./middlewares");
const openapi_context_1 = require("./framework/openapi.context");
const openapi_spec_loader_1 = require("./framework/openapi.spec.loader");
const util_1 = require("./middlewares/util");
const resolvers_1 = require("./resolvers");
const types_1 = require("./framework/types");
const res = require("./resolvers");
exports.default = openapiValidator;
exports.resolvers = res;
exports.middleware = openapiValidator;
exports.error = {
    InternalServerError: types_1.InternalServerError,
    UnsupportedMediaType: types_1.UnsupportedMediaType,
    RequestEntityToLarge: types_1.RequestEntityToLarge,
    BadRequest: types_1.BadRequest,
    MethodNotAllowed: types_1.MethodNotAllowed,
    NotFound: types_1.NotFound,
    Unauthorized: types_1.Unauthorized,
    Forbidden: types_1.Forbidden,
};
function openapiValidator(options) {
    const chainMiddleware = (handlers, req, res, next) => {
        let n = next;
        for (let i = handlers.length - 1; i >= 0; i--) {
            const c = handlers[i];
            const nxt = n;
            n = (err) => {
                if (err)
                    return next(err);
                else
                    c(req, res, nxt);
            };
        }
        n();
    };
    const oav = new OpenApiValidator(options);
    exports.middleware._oav = oav; // for testing
    const pspec = new openapi_spec_loader_1.OpenApiSpecLoader({
        apiDoc: options.apiSpec,
        $refParser: options.$refParser,
    }).load();
    return (req, res, next) => {
        pspec.then((spec) => {
            const context = new openapi_context_1.OpenApiContext(spec, oav.options.ignorePaths);
            oav.installPathParams(req.route || req.app, context);
            oav.installOperationHandlers(req.route || req.app, context);
            return spec;
        });
        chainMiddleware(oav.installMiddleware(pspec), req, res, next);
    };
}
class OpenApiValidator {
    constructor(options) {
        this.validateOptions(options);
        this.normalizeOptions(options);
        if (options.unknownFormats == null)
            options.unknownFormats === true;
        if (options.coerceTypes == null)
            options.coerceTypes = true;
        if (options.validateRequests == null)
            options.validateRequests = true;
        if (options.validateResponses == null)
            options.validateResponses = false;
        if (options.validateSecurity == null)
            options.validateSecurity = true;
        if (options.fileUploader == null)
            options.fileUploader = {};
        if (options.$refParser == null)
            options.$refParser = { mode: 'bundle' };
        if (options.validateFormats == null)
            options.validateFormats = 'fast';
        if (typeof options.operationHandlers === 'string') {
            /**
             * Internally, we want to convert this to a value typed OperationHandlerOptions.
             * In this way, we can treat the value as such when we go to install (rather than
             * re-interpreting it over and over).
             */
            options.operationHandlers = {
                basePath: options.operationHandlers,
                resolver: resolvers_1.defaultResolver,
            };
        }
        else if (typeof options.operationHandlers !== 'object') {
            // This covers cases where operationHandlers is null, undefined or false.
            options.operationHandlers = false;
        }
        if (options.validateResponses === true) {
            options.validateResponses = {
                removeAdditional: false,
            };
        }
        if (options.validateRequests === true) {
            options.validateRequests = {
                allowUnknownQueryParameters: false,
            };
        }
        if (options.validateSecurity === true) {
            options.validateSecurity = {};
        }
        this.options = options;
    }
    installMiddleware(spec) {
        const middlewares = [];
        const pContext = spec.then((spec) => new openapi_context_1.OpenApiContext(spec, this.options.ignorePaths));
        // metadata middleware
        middlewares.push((req, res, next) => pContext
            .then((context) => this.metadataMiddlware(context)(req, res, next))
            .catch(next));
        if (this.options.fileUploader) {
            // multipart middleware
            middlewares.push((req, res, next) => pContext
                .then((context) => this.multipartMiddleware(context)(req, res, next))
                .catch(next));
        }
        middlewares.push((req, res, next) => pContext
            .then((context) => {
            const components = context.apiDoc.components;
            if (this.options.validateSecurity && (components === null || components === void 0 ? void 0 : components.securitySchemes)) {
                return this.securityMiddleware(context)(req, res, next);
            }
            else {
                next();
            }
        })
            .catch(next));
        if (this.options.validateRequests) {
            middlewares.push((req, res, next) => {
                return pContext
                    .then((context) => this.requestValidationMiddleware(context)(req, res, next))
                    .catch(next);
            });
        }
        if (this.options.validateResponses) {
            middlewares.push((req, res, next) => pContext
                .then((context) => this.responseValidationMiddleware(context)(req, res, next))
                .catch(next));
        }
        return middlewares;
    }
    installPathParams(app, context) {
        const pathParams = [];
        for (const route of context.routes) {
            if (route.pathParams.length > 0) {
                pathParams.push(...route.pathParams);
            }
        }
        // install param on routes with paths
        for (const p of _uniq(pathParams)) {
            app.param(p, (req, res, next, value, name) => {
                const openapi = req.openapi;
                if (openapi === null || openapi === void 0 ? void 0 : openapi.pathParams) {
                    const { pathParams } = openapi;
                    // override path params
                    req.params[name] = pathParams[name] || req.params[name];
                }
                next();
            });
        }
    }
    metadataMiddlware(context) {
        return middlewares.applyOpenApiMetadata(context);
    }
    multipartMiddleware(context) {
        return middlewares.multipart(context, {
            multerOpts: this.options.fileUploader,
            unknownFormats: this.options.unknownFormats,
        });
    }
    securityMiddleware(context) {
        var _a;
        const securityHandlers = (_a = (this.options.validateSecurity)) === null || _a === void 0 ? void 0 : _a.handlers;
        return middlewares.security(context, securityHandlers);
    }
    requestValidationMiddleware(context) {
        const { coerceTypes, unknownFormats, validateRequests, validateFormats, } = this.options;
        const { allowUnknownQueryParameters } = (validateRequests);
        const requestValidator = new middlewares.RequestValidator(context.apiDoc, {
            nullable: true,
            coerceTypes,
            removeAdditional: false,
            useDefaults: true,
            unknownFormats,
            allowUnknownQueryParameters,
            format: validateFormats,
        });
        return (req, res, next) => requestValidator.validate(req, res, next);
    }
    responseValidationMiddleware(context) {
        const { coerceTypes, unknownFormats, validateResponses, validateFormats, } = this.options;
        const { removeAdditional } = validateResponses;
        return new middlewares.ResponseValidator(context.apiDoc, {
            nullable: true,
            coerceTypes,
            removeAdditional,
            unknownFormats,
            format: validateFormats,
        }).validate();
    }
    installOperationHandlers(app, context) {
        for (const route of context.routes) {
            const { method, expressRoute } = route;
            /**
             * This if-statement is here to "narrow" the type of options.operationHanlders
             * to OperationHandlerOptions (down from string | false | OperationHandlerOptions)
             * At this point of execution it _should_ be impossible for this to NOT be the correct
             * type as we re-assign during construction to verify this.
             */
            if (this.isOperationHandlerOptions(this.options.operationHandlers)) {
                const { basePath, resolver } = this.options.operationHandlers;
                app[method.toLowerCase()](expressRoute, resolver(basePath, route));
            }
        }
    }
    validateOptions(options) {
        if (!options.apiSpec)
            throw ono_1.default('apiSpec required');
        const securityHandlers = options.securityHandlers;
        if (securityHandlers != null) {
            if (typeof securityHandlers !== 'object' ||
                Array.isArray(securityHandlers)) {
                throw ono_1.default('securityHandlers must be an object or undefined');
            }
            util_1.deprecationWarning('securityHandlers is deprecated. Use validateSecurities.handlers instead.');
        }
        if (options.securityHandlers && options.validateSecurity) {
            throw ono_1.default('securityHandlers and validateSecurity may not be used together. Use validateSecurities.handlers to specify handlers.');
        }
        const multerOpts = options.multerOpts;
        if (multerOpts != null) {
            if (typeof multerOpts !== 'object' || Array.isArray(multerOpts)) {
                throw ono_1.default('multerOpts must be an object or undefined');
            }
            util_1.deprecationWarning('multerOpts is deprecated. Use fileUploader instead.');
        }
        if (options.multerOpts && options.fileUploader) {
            throw ono_1.default('multerOpts and fileUploader may not be used together. Use fileUploader to specify upload options.');
        }
        const unknownFormats = options.unknownFormats;
        if (typeof unknownFormats === 'boolean') {
            if (!unknownFormats) {
                throw ono_1.default("unknownFormats must contain an array of unknownFormats, 'ignore' or true");
            }
        }
        else if (typeof unknownFormats === 'string' &&
            unknownFormats !== 'ignore' &&
            !Array.isArray(unknownFormats))
            throw ono_1.default("unknownFormats must contain an array of unknownFormats, 'ignore' or true");
    }
    normalizeOptions(options) {
        // Modify the request
        if (options.securityHandlers) {
            options.validateSecurity = {
                handlers: options.securityHandlers,
            };
            delete options.securityHandlers;
        }
        if (options.multerOpts) {
            options.fileUploader = options.multerOpts;
            delete options.multerOpts;
        }
    }
    isOperationHandlerOptions(value) {
        if (value.resolver) {
            return true;
        }
        else {
            return false;
        }
    }
}
//# sourceMappingURL=index.js.map